""" Utility functions used by PXTL implementations.
"""

__all__= [
  'staticList', 'removeReparent', 'setDoctype', 'getNamespaces',
  'encodeText', 'decodeName', 'parsePseudoPIs', 'ConditionalContext',
  'checkFeatures', 'haveNestedScopes'
]

import types, sys, string, re
from pxtl.constants import *
from pxtl.exceptions import *


# DOM utility functions
#
def staticList(nodes):
  """ Turn a live DOM NodeList, NamedNodeMap or plain list into a non-live
      Python list containing the same nodes, which can be used for potentially
      destructive iteration.
  """
  if isinstance(nodes, types.ListType):
    return nodes[:]
  l= []
  for i in range(nodes.length):
    l.append(nodes.item(i))
  return l


def removeReparent(node):
  """ Replace a DOM node with its children, making sure all namespaces remain
      the same.
  """
  # Copy all non-overridden namespace directives from the node to all its
  # element children.
  #
  for i in range(node.attributes.length):
    attribute= node.attributes.item(i)
    if attribute.namespaceURI==NSNS:
      for child in staticList(node.childNodes):
        if child.nodeType==child.ELEMENT_NODE:
          if not child.hasAttributeNS(NSNS, attribute.localName):
            child.setAttributeNS(NSNS, attribute.name, attribute.value)

  # Replace node with a DocumentFragment of all its children.
  #
  frag= node.ownerDocument.createDocumentFragment()
  for child in staticList(node.childNodes):
    frag.appendChild(child)
  node.parentNode.replaceChild(frag, node)

def setDoctype(document, doctype):
  """ Try to set the doctype of a document to the specified value. DOM Level
      2 implementations and some Level 3 implementations don't support this;
      in this case, clone the document's content into a new document with the
      correct doctype.
  """
  current= document.doctype
  try:
    if current is None:
      if doctype is None:
        return document
      else:
        document.insertBefore(doctype, document.firstChild)
    else:
      if doctype is None:
        document.removeChild(current)
      else:
        document.replaceChild(doctype, current)
    return document
  except Exception:

    # Failed in some way (why can't tell what, since the exception could be
    # anything). Try the copying method instead.
    #
    imp= document.implementation
    newDocument= imp.createDocument(PXNS, 'px:temp', doctype)
    frag= newDocument.createDocumentFragment()
    for child in staticList(document.childNodes):
      if child.nodeType!=child.DOCUMENT_TYPE_NODE:
        frag.appendChild(newDocument.importNode(child, True))
    newDocument.replaceChild(frag, newDocument.documentElement)

    # Try to copy any of the new properties from DOM Level 3 to the new
    # document; any of this could be unsupported and fail
    #
    try:
      newDocument.xmlVersion= document.xmlVersion
    except AttributeError:
      pass
    try:
      newDocument.xmlEncoding= document.xmlEncoding
    except AttributeError:
      pass
    try:
      newDocument.xmlStandalone= document.xmlStandalone
    except AttributeError:
      pass
    try:
      newDocument.documentURI= document.documentURI
    except AttributeError:
      pass

    return newDocument


def getNamespaces(node):
  """ Get a mapping {prefix: uri} of all non-built-in namespaces active on an
      element.
  """
  nss= {}
  while node.nodeType==node.ELEMENT_NODE:
    for attr in staticList(node.attributes):
      if attr.namespaceURI==NSNS:
        prefix= attr.localName
        if attr.localName=='xmlns':
          prefix= None
        if not nss.has_key(prefix):
          nss[prefix]= attr.value
    node= node.parentNode
  return nss


# General-purpose utility functions.
#
class DictAsInstance:
  def __init__(self, d):
    self.__dict__= d

_encode_upar= re.compile('[^a-zA-Z0-9\\-_\\.]')
_encode_jstr= re.compile('[\\\'"&<>-]')
_encode_cstr= _encode_jstr
_encode_name= re.compile('[^a-zA-Z]')
_decode_name= re.compile('_..')

_hexdigits= '0123456789ABCDEF0123456789abcdef'
def _hexbyte(x):
  c= ord(x)
  return _hexdigits[int(c/16)]+_hexdigits[c%16]


def encodeText(text, coding):
  """ Return text in an escaped form (for URL parameters, JS literals etc.).
  """
  if coding in ('upar', 'name'):
    if type(text)==UnicodeType:
      text= text.encode('utf-8')
  if coding=='text':
    return text
  elif coding=='upar':
    return _encode_upar.sub(lambda m: '%'+_hexbyte(m.group(0)), text)
  elif coding=='jstr':
    return _encode_jstr.sub(lambda m: '\\x'+_hexbyte(m.group(0)), text)
  elif coding=='cstr':
    return _encode_cstr.sub(lambda m: '\\x'+_hexbyte(m.group(0))+' ', text)
  elif coding=='name':
    return _encode_name.sub(lambda m: '_'+_hexbyte(m.group(0)), text)
  else:
    raise ValueError ("unknown coding '%s'" % coding)


def decodeName(text):
  text= _decode_name.sub(lambda m: chr(eval('0x'+m.group(0)[1:])), text)
  if UnicodeType is None:
    return text
  return unicode(text, 'utf-8')


_ws= [' ', '\n', '\t']
def parsePseudoPIs(attr):
  """ Break an attribute value up into text and pseudo-PI sections
      (target, data).
  """
  parts= []
  ix= 0
  while True:
    ox= string.find(attr.value, '{?', ix)
    if ox==-1:
      parts.append(attr.value[ix:])
      break
    parts.append(attr.value[ix:ox])
    cx= string.find(attr.value, '?}', ox+1)
    if cx==-1:
      raise PXTLPseudoPIOpenError(attr, ox)
    if cx==ox+1:
      parts.append('{?')
    else:
      wx= ox+2
      while wx<cx:
        if attr.value[wx] in _ws:
          break
        wx= wx+1
      vx= wx
      while vx<cx:
        if attr.value[vx] not in _ws:
          break
        vx= vx+1
      parts.append((attr.value[ox+2:wx], attr.value[vx:cx]))
    ix= cx+2
  return parts


def checkFeatures(future):
  """ Convert a future attribute node into a list of optional extra features.
      Raise an exception for any features that cannot be used in the current
      version of Python.
  """
  if future==None:
    return []
  features= map(string.strip, string.split(future.value, ','))
  for feature in features[:]:

    # If the feature is not implemented at all in the current version of
    # Python, complain. If the feature is on by default anyway, don't bother
    # including it in the import list.
    #
    if hasattr(future__mod, feature):
      f= getattr(future__mod, feature)
      if sys.version_info>=f.getOptionalRelease():
        if sys.version_info>=f.getMandatoryRelease():
          features.remove(feature)
        continue
    raise PXTLFutureFeatureError(feature, future)
  return features


def haveNestedScopes(features):
  """ Determine whether PXTL should provide nested scopes to subtemplate
      functions. Check the supplied features string and the current Python
      version's default features.
  """
  if 'nested_scopes' in features:
    return True
  if hasattr(future__mod, 'nested_scopes'):
    if sys.version_info>=future__mod.nested_scopes.getMandatoryRelease():
      return True
  return False


class ConditionalContext:
  """ Keeps track of previous clauses of an if-orif-anif-elif-else construct and
      calculates the success of each element. Used by conditional elements and
      conditional attributes.
  """
  def reset(self):
    """ Start a new conditional context, forgetting any previous clauses. This
        also serves as the constructor.
    """
    self.old_cond= None
    self.old_testing= None
    self.old_success= None
  __init__= reset

  def getTesting(self, cond):
    """ Given the new clause 'cond' and knowledge of previous clauses, decide
        whether the clause's test condition needs to be evaluated to determine
        its success. If not, laziness kicks in and we can avoid doing the test.
        getTesting for 'else' may return True even though 'else' doesn't
        actually have a clause to evaluate. If the clause ordering makes no
        sense (eg. if-else-orif), return None instead of a boolean.
    """
    if cond=='if':
      return True
    elif cond=='orif':
      if self.old_cond in ('if', 'orif'):
        return not self.old_success
    elif cond=='anif':
      if self.old_cond in ('if', 'orif', 'anif', 'for', 'while'):
        return self.old_success
    elif cond in ('elif', 'else'):
      if self.old_cond in ('if', 'orif', 'anif', 'for', 'while'):
        return not self.old_success
      if self.old_cond=='elif':
        return not self.old_success and self.old_testing
    return None

  def getSuccess(self, cond, test):
    """ Work out whether a clause is successful, given the truth of its
        evaluated test condition. If the condition was not evaluated (due to
        getTesting having returned False), 'test' should be None instead of a
        boolean. Store the clause info as it may affect a following clause.
    """
    if cond=='orif':
      success= test in (None, True)
    else:
      success= test
    self.old_cond= cond
    self.old_testing= test!=None
    self.old_success= success
    return success

  def copy(self):
    """ Make a copy of a ConditionalContext. Used for conditional attributes
        which can be hierarchically nested.
    """
    clone= ConditionalContext()
    clone.old_cond= self.old_cond
    clone.old_testing= self.old_testing
    clone.old_success= self.old_success
    return clone
